import numpy as np
import pandas as pd
from scipy import linalg, stats
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns
import click


def generate_state_var(S):
    return [(k, m) for m in range(S + 1) for k in range(S + 1) if k + m <= S]


def generate_rate_matrix(S, lam, mu, gamma, stateVar=None):
    if stateVar is None:
        stateVar = generate_state_var(S)

    RateMatrix = np.zeros((len(stateVar), len(stateVar)))

    for down, (k_down, m_down) in enumerate(stateVar):
        for across, (k, m) in enumerate(stateVar):
            if k == k_down - 1 and m == m_down:
                RateMatrix[down, across] = (S - m - k) * (k * lam / (S - 1) + 2 * mu)
            elif k == k_down and m == m_down - 1:
                RateMatrix[down, across] = m * (S - m - k) * lam / (S - 1)
            elif k == k_down + 1 and m == m_down - 1:
                RateMatrix[down, across] = k * (m * lam / (S - 1) + mu)
            elif k == k_down + 1 and m == m_down:
                RateMatrix[down, across] = k * ((S - m - k) * lam / (S - 1) + gamma)
            elif k == k_down and m == m_down + 1:
                RateMatrix[down, across] = m * (S - m - k) * lam / (S - 1)
            elif k == k_down - 1 and m == m_down + 1:
                RateMatrix[down, across] = m * (k * lam / (S - 1) + 2 * gamma)
            elif k == k_down and m == m_down:
                RateMatrix[down, across] = -(2 * ((k + m) * (S - m - k) + k * m) * lam / (S - 1) + (k + 2 * m) * gamma + (2 * S - (k + 2 * m)) * mu)

    return RateMatrix


def findProbDist(RateMatrix, InitialConditions, stateVar, age, S):

    ProbStates = linalg.expm(RateMatrix * age) @ InitialConditions
    ProbDist = np.zeros(2 * S + 1)
    for index, (k, m) in enumerate(stateVar):
        ProbDist[k + 2 * m] += ProbStates[index]

    return ProbDist


def runModel(S, lam, mu, gamma, age):

    stateVar = generate_state_var(S)
    RateMatrix = generate_rate_matrix(S, lam, mu, gamma, stateVar)
    InitialConditions = np.zeros(int(0.5 * (S + 1) * (S + 2)))
    InitialConditions[[0, -1]] = 0.5
    ProbDist = findProbDist(RateMatrix, InitialConditions, stateVar, age, S)

    return ProbDist


def beta_convert_params(mu, kappa):
    """
    Convert mean/dispersion parameterization of a beta distribution to the ones scipy supports

    """

    if np.any(kappa <= 0):
        raise Exception("kappa must be greater than 0")
    elif np.any(mu <= 0) or np.any(mu >= 1):
        raise Exception("mu must be between 0 and 1")

    alpha = kappa * mu
    beta = kappa * (1 - mu)

    return alpha, beta


def rescale_beta(beta, delta, eta):
    # Linear transform of beta values from between
    # 0 and 1 to between delta and eta
    return (eta - delta) * beta + delta


def add_noise(beta, delta, eta, kappa):
    beta_rescale = rescale_beta(beta, delta, eta)
    a, b = beta_convert_params(beta_rescale, kappa)

    return stats.beta.rvs(a, b)


def simulate_beta(params, S, age, N):
    lam, mu, gamma, delta, eta, kappa = params

    ProbDist = runModel(S, lam, mu, gamma, age)

    k_sample = np.random.choice(np.arange(0, 2 * S + 1), size=N, p=ProbDist)
    beta_sample = k_sample / (2 * S)

    return add_noise(beta_sample, delta, eta, kappa)


@click.command()
@click.option("--betas", help="Original csv file with the beta files", required=True, type=click.Path(exists=True))
@click.option("--log", help="BEAST log output", required=True, type=click.Path(exists=True))
@click.option("--samplesheet", help="Metadata CSV containing sample and age columns", required=True, type=click.Path(exists=True))
@click.option('--output', type=click.Path(exists=False), required=True, help='Output name (pdf)')
@click.option("--cells", "S", help="Number of stem cells used in the model", required=True, type=int)
@click.option("--sample_col", help="Number of stem cells used in the model", required=True, type=str)
@click.option("--age_col", help="Number of stem cells used in the model", required=True, type=str)
def main(betas, log, samplesheet, output, S, sample_col, age_col):

    assert samplesheet.endswith(".csv") and betas.endswith(".csv") and log.endswith(".log"), "Unexpected input format. The betas and samplesheet should be a .csv file, while log should be a .log file"
    df = pd.read_csv(betas, index_col=0)
    df.columns = df.columns.astype(str)

    posterior = pd.read_csv(log, index_col=0, sep="\t", comment="#")
    posterior = posterior[["flipflop.lambda", "flipflop.mu", "flipflop.gamma", "errorModel.deltaOffset", "errorModel.etaOffset", "errorModel.kappaScale"]]

    samplesheet = pd.read_csv(samplesheet)
    ages = samplesheet.groupby(sample_col)[age_col].max().to_dict()

    with PdfPages(output) as export_pdf:
        for sample in df.columns:
            y = df[sample]
            N = y.shape[0]
            P = posterior.shape[0]
            age = ages[sample]
            
            y_hat = np.empty((P, N))

            for p in range(P):
                params = posterior.iloc[p].values
                y_hat[p, :] = simulate_beta(params, S, age, N)

            fig, ax = plt.subplots()
            plt.hist(y, np.linspace(0, 1, 101), density=True, alpha=0.4, linewidth=0)
            plt.hist(np.ravel(y_hat), np.linspace(0, 1, 101), density=True, alpha=0.4, linewidth=0)
            plt.legend(("Data", "Posterior Predictive"))
            plt.xlabel("Fraction Methylated (Beta)")
            plt.ylabel("Probability Density")
            plt.title(f"Sample ID: {sample}")
            sns.despine()
            plt.tight_layout()
            export_pdf.savefig(dpi=600)
            plt.close()


if __name__ == "__main__":
    main()



import pytest
from pathlib import Path


@pytest.fixture()
def example_config():
    return Path("tests/data/config.yaml")


@pytest.fixture()
def defaults():
    return Path("src/phyfum/workflow/config.schema.yml")



import pytest
from pathlib import Path
import snakemake
import src.phyfum.workflow.aux.checkConfig as checkConfig


def test_defaults(defaults: Path):
    cleanConfig = checkConfig.readDefaults(defaults)

    assert isinstance(cleanConfig, dict)
    assert len(cleanConfig) > 0


def test_config_cleanup(defaults: Path, example_config: Path):
    DEFAULTS = checkConfig.readDefaults(defaults)
    config = snakemake.common.configfile.load_configfile(example_config)
    finalConfig = checkConfig.makeConfig(config, DEFAULTS)
    assert len(finalConfig) > 3 and finalConfig.get("mle_iterations") == 7500



# SPDX-FileCopyrightText: 2023-present pbousquets <pbousquetsm@gmail.com>
#
# SPDX-License-Identifier: MIT
import sys

if __name__ == "__main__":
    from phyfum.cli import phyfum

    sys.exit(phyfum())



import pandas as pd
from typing import List


def checkPatientsBetas(patients: List[str], dir: str) -> List[str]:
    validPatients = []
    for patient in patients:
        csv_file = f"{dir}/{patient}.csv"

        df = pd.read_csv(csv_file)
        if len(df.columns) >= 3:
            validPatients.append(patient)
    if len(validPatients) == 0:
        raise ValueError(f"No patients with at least 3 samples were found! Please check your input data. {df.columns}")

    return validPatients


def checkPatientInfo(patientInfo, patient_col):
    patientInfo = patientInfo[patientInfo.groupby(patient_col)[patient_col].transform('count') >= 3]
    if len(patientInfo) == 0:
        raise ValueError("No patients with at least 3 samples were found! Please check your input data.")
    return patientInfo



from os.path import isdir
import re
import glob


def checkInput(config: dict) -> str:
    if config.get("input").lower().endswith("csv"):
        input_type = "trees"
    elif isdir(config.get("input")):
        folder = config.get("input")
        sheet = config.get("patientInfo", "NoSheetFound")
        files = glob.glob(f'{folder}/**/*', recursive=True)

        grn_idat_pattern = re.compile(r'Grn\.idat', re.IGNORECASE)
        red_idat_pattern = re.compile(r'Red\.idat', re.IGNORECASE)
        samplesheet_pattern = re.compile(rf'{sheet}', re.IGNORECASE)

        grn_idat_found = any(grn_idat_pattern.search(file) for file in files)
        red_idat_found = any(red_idat_pattern.search(file) for file in files)
        samplesheet_found = any(samplesheet_pattern.search(file) for file in files)

        if grn_idat_found and red_idat_found and samplesheet_found:
            input_type = "complete"
        elif grn_idat_found and red_idat_found and not samplesheet_found:
            raise RuntimeError("IDAT files detected, but the samplesheet couldn't be found in the input folder. Check that the file contains the 'samplesheet' keyword in its name and that it's a csv file.")
        else:
            raise RuntimeError(
                "Unexpected input folder. Check that the folder contains IDAT files for both red (Red.idat) and green (Grn.idat) and a samplesheet. Check that parameter --pattern matches your samplesheet. If the issue persists, contact the developers."
            )
    else:
        raise RuntimeError(
            """Couldn't determine the input type.
        - If running a single beta file, make sure the file is in csv format and ends with '.csv'
        - If running a folder of multiple csv files (multiple individuals), make sure csv files are in the folder 
        - If running the complete workflow, make sure a the samplesheet is in the folder (the file name should contain 'samplesheet') and that IDAT files are located recursively within the folder. 
        """
        )
    return input_type



import pandas as pd
from typing import List


class methylationSample:
    """
    Class for reading samples and casting them into a single comma-separated element
    """

    def __init__(self, methylationArray: pd.Series, precision: int):
        self.methylationArray = methylationArray.round(precision)  # Round floats, we don't need long decimals

    def toString(self) -> str:
        return self.methylationArray.astype("string").str.cat(sep=",")  # Join all sites into a single comma-separated string


class readMethylation:
    """
    Class representing the whole dataset
    """

    def __init__(self, inputStr: str, precision: int, stripRownames: bool, samplesheet: str, age_col: str, age_diagnosis_col: str, sample_col: str, sample_type_col:str, luca_mode: str):
        index_col = 0 if stripRownames else None
        self.input = pd.read_csv(inputStr, index_col=index_col)
        self.precision = precision
        self.luca_mode = luca_mode
        self.samples: List[str] = []
        self.sample_names: List[str] = self.input.columns.tolist()
        self.checkValues()
        self.parseSampleSheet(samplesheet, age_col, age_diagnosis_col, sample_col, sample_type_col)

    def parseSamples(self) -> None:
        for column in self.sample_names:
            sample: methylationSample = methylationSample(self.input[column], self.precision)
            self.samples.append(sample.toString())

    def checkValues(self) -> None:
        assert len(self.input.select_dtypes(include="object").columns) == 0, "\nStrings found in the dataframe. Did you to forget using the --stripRownames flag?\n"
        assert all(self.input.max() <= 1) and all(self.input.min()) >= 0, "\nThe data frame contains values not in (0,1) range. \nDid you to forget using the --stripRownames flag?\n"

    def parseSampleSheet(self, samplesheet: str, age_col: str, age_diagnosis_col: str, sample_col: str, sample_type_col: str) -> None:
        self.samplesheet = pd.read_csv(samplesheet)
        assert sample_col in self.samplesheet.columns, f"Column {sample_col} to estimate samples not found in sample sheet"
        assert age_col in self.samplesheet.columns, f"Column {age_col} to estimate ages not found in sample sheet"
        assert sample_type_col in self.samplesheet.columns, f"Column {sample_type_col} to estimate sample types not found in sample sheet"

        if age_diagnosis_col:
            assert age_diagnosis_col in self.samplesheet.columns, f"Column {age_diagnosis_col} not found in sample sheet"

        # Filter samples with sample_names
        self.samplesheet = self.samplesheet[self.samplesheet[sample_col].isin(self.sample_names)]

        # Check if there are normals
        self.normals = self.samplesheet[sample_type_col].str.lower().str.count('normal|control').sum() > 0

        self.fixed_luca = True if self.luca_mode == "fixed" or (self.luca_mode.lower() == "auto" and self.normals) else False
        # If present, age_diagnosis is the same across all samples. Check it and save it as an int attribute
        if age_diagnosis_col:
            assert self.samplesheet[age_diagnosis_col].nunique() == 1, "Age at diagnosis is not the same across all samples"
            self.age_diagnosis = int(self.samplesheet[age_diagnosis_col].unique()[0])
        else: # If not present, age_diagnosis is the youngest sample in age_col
            self.age_diagnosis = self.samplesheet[age_col].min()
        
        # DOLS is the age of the oldest sample in age_col, DOFS is the age of the first sample
        self.age_dols = self.samplesheet[age_col].max()
        self.age_dofs = self.samplesheet[age_col].min()

        # Save age as a dict of sample_name: age
        self.ages = self.samplesheet.set_index(sample_col)[age_col].to_dict()
        self.tree_settings = {
            "heightRules": "true" if self.fixed_luca else "false",
            "cenancestorHeight": {
                "value": str(self.age_dols) if self.fixed_luca else str(self.age_dols - self.age_diagnosis),
            },
            "cenancestorBranch": {
                "value": "1",
                "lower": "0",
                "upper": str(self.age_dofs),
            },
            "fixed_luca": self.fixed_luca
        }

        if not self.fixed_luca:
            self.tree_settings["cenancestorHeight"]["lower"] = str(self.age_dols - self.age_diagnosis)
            self.tree_settings["cenancestorHeight"]["upper"] = str(self.age_dols)
            self.tree_settings["luca_height_prior"] = {
                "lower": str(self.age_dols - self.age_diagnosis),
                "upper": str(self.age_dols)
                }



import yaml
from os.path import abspath


def readDefaults(yaml_path: str) -> dict:
    with open(yaml_path, 'r') as f:
        config = yaml.safe_load(f)
    default_values = {key: value["default"] for key, value in config.get("properties", {}).items() if "default" in value}
    return default_values


def makeConfig(config: dict, defaults: dict) -> dict:
    """
    Remove Nones from config and add missing variables from the defaults
    """
    config = {key: value for key, value in config.items() if value is not None and value != "None"}

    # Merge 'config' with 'defaults' to add missing variables
    config = {**defaults, **config}
    if not config.get("mle_iterations"):
        config["mle_iterations"] = int(config.get("iterations") / config.get("mle_steps"))
    if not config.get("mle_sampling"):
        config["mle_sampling"] = int(config.get("iterations") / 1000)
    if config.get("blacklist"):
        config["blacklist"] = abspath(config.get("blacklist"))
    return config



#!/usr/bin/env python3

import pandas as pd
import numpy as np
import argparse


def filter_noise(beta, M, U, r_min=500):
    fil_index = M + U >= r_min

    if len(np.shape(beta)) > 1:
        fil_index = fil_index.all(axis=1)

    M_fil = M[fil_index]
    U_fil = U[fil_index]
    beta_fil = beta[fil_index]

    return beta_fil, M_fil, U_fil


def subset_sites(beta, M, U, samples, patients, patientinfo, patient_col, p=5):

    patientinfo = patientinfo.loc[samples]

    beta_fil, M_fil, U_fil = filter_noise(beta, M, U, r_min=1200)

    mean_beta = np.mean(beta_fil, axis=1)

    patient_std = pd.DataFrame({})
    for patient in patients:
        patient_samples = list(patientinfo.loc[patientinfo[patient_col] == patient, :].index)
        patient_std[patient] = np.std(beta_fil[patient_samples], axis=1)

    mean_std = np.mean(patient_std, axis=1)

    thresh = np.percentile(mean_std, 100 - p)
    index = (mean_std > thresh) & (0.4 < mean_beta) & (mean_beta < 0.6)

    return index[index == True].index


def main():
    parser = argparse.ArgumentParser(description="Select tick-tock CpG Loci")
    parser.add_argument("outputfile", type=str, help="path to csv file in which to store output")
    parser.add_argument("betafile", type=str, help="path to csv file containing noob beta values")
    parser.add_argument("methylatedfile", type=str, help="path to csv file containing noob M values")
    parser.add_argument("unmethylatedfile", type=str, help="path to csv file containing noob U values")
    parser.add_argument(
        "patientinfofile",
        type=str,
        help="""
                        path to csv file containing patient details (must 
                        contain sample IDs that are the columns of betafile
                        and a column titled "patient")
                        """,
    )
    parser.add_argument("-e", "--manifestfile", required=True, help="path to Illumina EPIC manifest csv file (default MethylationEPIC_v-1-0_B4.csv)")
    parser.add_argument("-c", "--crossreactivefile", required=True, help="path to csv file containing CpGs to exclude (default 13059_2016_1066_MOESM1_ESM.csv)")
    parser.add_argument("-b", "--basename", action="store_true", help="Use basename from EPIC sheet to infere the sample ")
    parser.add_argument("-P", "--patient_col", default="Patient", type=str, help="Name of the patient column in the patientinfofile (default Patient)")
    parser.add_argument("-S", "--sample_col", default="Sample", type=str, help="Name of the sample column in the patientinfofile (default Patient)")
    parser.add_argument(
        "-t",
        "--tissue",
        default=None,
        help="""
                        str determining which samples to include - if not None then 
                        patientinfo must contain a column titled "location" 
                        containing at least one entry corresponding to this option
                        (default None)
                        """,
    )

    parser.add_argument(
        "-d",
        "--disease",
        default=None,
        help="""
                        str determining which samples to include - if not None then 
                        patientinfo must contain a column titled "disease" 
                        containing at least one entry corresponding to this option
                        (default None)
                        """,
    )
    parser.add_argument(
        "-p",
        "--percent",
        default=5,
        help="""
                        float specifying the top p%% of most heterogeneous CpG loci
                        to include 
                        """,
    )
    # Execute the parse_args() method
    args = parser.parse_args()

    outputfile = args.outputfile
    betafile = args.betafile
    methylatedfile = args.methylatedfile
    unmethylatedfile = args.unmethylatedfile
    patientinfofile = args.patientinfofile
    manifestfile = args.manifestfile
    crossreactivefile = args.crossreactivefile
    tissue = args.tissue
    disease = args.disease
    patient_col = args.patient_col
    sample_col = args.sample_col

    p = float(args.percent)

    if (p <= 0) | (p >= 100):
        raise Exception("p must be number between 0 and 100")

    beta_all = pd.read_csv(betafile, index_col=0)
    M_all = pd.read_csv(methylatedfile, index_col=0)
    U_all = pd.read_csv(unmethylatedfile, index_col=0)

    patientinfo = pd.read_csv(patientinfofile, keep_default_na=False, index_col=False)
    indexes = patientinfo[sample_col]
    patientinfo.index = indexes

    crossreactive = pd.read_csv(crossreactivefile, index_col=0)
    manifest = pd.read_csv(manifestfile, index_col=0, low_memory=False)

    difference = beta_all.index.difference(crossreactive.index)

    beta_all = beta_all.loc[difference]
    M_all = M_all.loc[difference]
    U_all = U_all.loc[difference]

    manifest = manifest.loc[manifest.index.intersection(beta_all.index), :]

    good_sites = manifest.loc[(manifest["Relation_to_UCSC_CpG_Island"].isna() & ~(manifest["CHR"].str.contains("X|Y")) & (manifest["Infinium_Design_Type"] == "II")), :].index

    if tissue is None and disease is None:
        samples = patientinfo.index
    elif disease is None:
        samples = patientinfo.loc[patientinfo["location"] == tissue, :].index
    elif tissue is None:
        samples = patientinfo.loc[patientinfo["disease"] == disease, :].index
    else:
        samples = patientinfo.loc[(patientinfo["disease"] == disease) & (patientinfo["location"] == tissue), :].index

    patients = pd.unique(patientinfo.loc[samples, patient_col])
    beta_loc = beta_all.loc[good_sites, samples]
    M_loc = M_all.loc[good_sites, samples]
    U_loc = U_all.loc[good_sites, samples]

    ticktock_index = subset_sites(beta_loc, M_loc, U_loc, samples, patients, patientinfo, patient_col, p)

    beta = beta_all.loc[ticktock_index, samples]

    beta.to_csv(outputfile)


if __name__ == "__main__":
    main()
